/*
	countdown.cpp
	Copyright Risto Holopainen 2021
	
	Generator of the composition The Final Countdown
	dual, triple fm, parametric seq/loop

	Note: different choices of optimisation flags may produce different 
	output, see comments in fm23.h
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.*/


#include "fm23.h"
#include "parleur.h"
#include <cstdio>
#include <sndfile.h>



// map features --> parameters

void tune3c(double cf[3])
{
	double a=cf[0], b=cf[1], c=cf[2];
	double s = a+b+c;
	double h0=0.012, h1=0.013, h2=0.015;
	if(s < 3600)
		{
		a += h0;
		b += h1;
		c += h2;
		}
	if(c > 2500)
		c -= 1.5*b;
	if(b > 2400)
		b -= 1.1*a;
	if(a > 2300)
		a *= 0.02;
	cf[0] = a + h0;
	cf[1] = b + h1;
	cf[2] = c + h2;
}


void tune3m(double mf[3])
{
	for(int k=0; k<3; k++)
		if( (mf[k] *= 0.999875) < 35+3*k)
			mf[k] += 700 + 7*k;
}


void retune_par3(const features &X, triple_fm &tf)
{
	double j = (X.amp + 0.0015) * 8.5;
	double k = (X.var + X.zc) * 0.002 + 0.0001;
	k += X.amp1k;
	tf.set_J(j);
	tf.set_K(k);
	printf("a: %.2f, 1k: %.3f, zc: %.3f, j: %.3f, k: %.3f \n", 
		X.amp, X.amp1k, X.zc, j, k);
}


void reset_frq3(triple_fm &tf)
{
	// tune osc unison
	double carr[3], mods[3], avg_c, avg_m, sum_c=0, sum_m=0;
	tf.get_carriers(carr);
	tf.get_modulators(mods);
	
	for(int i=0; i<3; i++)
		{
		sum_c += carr[i];
		sum_m += mods[i];
		}
	avg_c = sum_c / 3;
	avg_m = sum_m / 3;
	
	printf("reset_f3 c: %.1f Hz, m: %.1f Hz \n", avg_c, avg_m);
	
	for(int i=0; i<3; i++)
		{
		carr[i] = .52*avg_c + .46*avg_m + .01*carr[i];
		mods[i] = .52*avg_m + .46*avg_c + .01*mods[i];	
		}
	tf.set_cf(carr);
	tf.set_mf(mods);
}


double phase_inject(const stereo &s)
{
	double phi = 0.25 * (fabs(s.L) + fabs(s.R));
	return phi;
}


void update_simple(simple_fm &sfm, const double cc[3], const double md[3], double indx, double fb)
{
	static int m=0, n=0, t=0;
	if(++t == 7)
		{
		++n %= 3;
		t = 0;
		}
	++m %= 3;

	double cfrq = cc[m];
	double mfrq = md[n];
	
	double r = cfrq / mfrq;
	
	if (r < 1/2.) r = 1/2.;
	else
		r = ((int) (4*r)) / 4.;
	
	mfrq = cfrq / r;
	sfm.set(cfrq, mfrq, indx, fb);
}

void init_dual(params &pm)
{
	pm.J = 0.3; 
	pm.K = 0.1;
	pm.A = 0.9;
	pm.M = 0.5;
	pm.carr_1 = 475;
	pm.carr_2 = 500;
	pm.omega_1 = 360;
	pm.omega_2 = 375;
}


void tune_dual(params &pm)
{
	double cc = .5*(pm.carr_1 + pm.carr_2);
	double mm = .5*(pm.omega_1 + pm.omega_2);
	
	pm.carr_1 = cc + 1;
	pm.carr_2 = cc - 1;
	pm.omega_1 = pm.omega_2 = mm;	
	if(pm.M > 0.6)
		pm.M *= 0.2;
	if(pm.K > 0.8)
		pm.K *= 0.1;
	if(pm.J > 1.2)
		pm.J = .5*pm.K;
}

void nudge_dual(dual_fm &df)
{
	params p = df.get_params();
	p.carr_1 *= 1.01;
	p.carr_2 *= 0.99;
	p.omega_1 += 0.2;
	p.omega_2 -= 0.1;
	
	p.K += 0.002;
	if(p.K > 0.6)
		p.K *= 0.05;
	p.J -= 0.005;
	if(p.J < 0.01)
		p.J = p.M;
	p.M += 0.02*p.K;	
	df.set(p);
}



bool sometimes(const features &phi)
{
	static int t = 0;
	if(1./phi.var > 10*phi.amp)
		++t;
	
	bool trig = false;
	if(t > 60 && phi.zc < 0.04)
		{
		trig = true;
		t = 0;
		}
	
	return trig;
}

void modify_eg(envelope &eg, const features &Phi)
{
	// adaptive prob.
	static int type[3] = {0,0,0};
	static double t1 = 0.165, t2 = 0.155;
	const int p=5;
	const double eps = 0.003;
	
	eg.duty(.6 - Phi.amp);
	if (Phi.amp > t1)
		{
		eg.shape(shapes::RAMP);
		++type[0];
		}
	else if (Phi.amp < t2)
		{
		eg.shape(shapes::EXPO);
		++type[1];
		}
	else
		{
		eg.shape(shapes::SQUARE);	
		++type[2];
		}
	
	if(type[0] > type[1]+p && type[0] > type[2]+p && t1 < 0.25)
		t1 += eps;
	else if(type[0] < type[1]-p && type[0] < type[2]-p && t1 > 0.05)
		t1 -= eps;
	
	
	if(type[1] > type[0]+p && type[1] > type[2]+p && t2 > 0.05)
		t2 -= eps;
	else if(type[1] < type[0]-p && type[1] < type[2]-p && t2 < 0.25)
		t1 += eps;
}


double squish_env(double a, double b)
{
	static double A = 0, y = 0;
	A = .5*(y + A);
	y = tanh(a + b + A);
	return y;
}

void set_period(envelope &eg, int k)
{
	// k: counter
	double cc = eg.get_cycle();
	if(k == 900)
		eg.cycle(7.5);
	else if(k == 1500)
		eg.cycle(2.4);
	else if(k > 2000 && cc < 12.0)
		{
		cc += 0.125;
		eg.cycle(cc);	
		}
}

int main()
{
	const int sr = 48000;
	params pmd;
	dual_fm DAF(sr);
	init_dual(pmd);
	DAF.set(pmd);
	triple_fm TAF(sr);
	simple_fm SAF(sr, 311, 365);


	compenv hack(sr);
	looper boucle_1(1750);
	boucle_1.set_loop(150,1660);
	looper boucle_2(1500);
	boucle_2.set_loop(200,1444);
	bool recording = false;
	double f1 = 2, f2 = 3;
	double x;
	stereo s2{0,0}, s3{0,0};	
	const int L = 1 << 5;
	const int N = 395*sr/L;
	double buf2[2*L], buf3[2*L], bufx[L];
	
	SNDFILE *ouf2, *ouf3, *oufx;
	SF_INFO head, monohead;
	head.samplerate = sr;
	head.channels = 2;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	monohead = head; monohead.channels = 1;
	
	ouf2 = sf_open("countdown_s2.wav", SFM_WRITE, &head);
	ouf3 = sf_open("countdown_s3.wav", SFM_WRITE, &head);
	oufx = sf_open("countdown_mx.wav", SFM_WRITE, &monohead);
	

	TAF.set_J(0.5);
	double cf_triple[3] = {100,99,100};
	double mf_triple[3] = {100,99,101};
	
	analyzer ANX1(sr), ANX2(sr);
	features Phi1, Phi2;	
	transient trans1, trans2;
	envelope Env(sr, EXPO, 3.6, 0.4);
	double s_amp = 0, d_amp = 0, t_amp = 0;
	double zz = 0;
	int somecounter = 0;

	for(int i=N; i>0; --i)
		{
		double emm = 0.01*f1;
		TAF.set_M( .5*emm*emm );
		if(i%17500 == 1)
			printf("M: %.4f \n", emm);
		TAF.get_carriers(cf_triple);
		TAF.get_modulators(mf_triple);
		tune3c(cf_triple);
		tune3m(mf_triple);
		TAF.set_cf(cf_triple);
		TAF.set_mf(mf_triple);
		
		for(int j=0, k=0; j<2*L; j += 2, k++)
			{
			s2 = DAF.gen();
			d_amp = hack.env();
			
			x = SAF.gen( phase_inject(s3) );
			s_amp = Env.env();
			t_amp = squish_env(1-s_amp, 2*d_amp);
			x *= s_amp;
			s3 = TAF.gen();
			
			buf2[j  ] = .5*d_amp*s2.L;
			buf2[j+1] = .5*d_amp*s2.R;
			buf3[j  ] = .5*t_amp*s3.L;
			buf3[j+1] = .5*t_amp*s3.R;
			bufx[k] = .5*x;

			Phi1 = ANX1(s3.L);
			Phi2 = ANX2(s3.R);
			if ( trans1.detect(Phi2) )
				{
				retune_par3(Phi1, TAF);
				modify_eg(Env, Phi1);
				
				}
			if ( trans2.detect(Phi1) )
				{
				reset_frq3(TAF);
				double fbk = 3.0*(Phi2.amp - Phi1.amp1k);
				update_simple(SAF, cf_triple, mf_triple, emm, fbk);
				tune_dual(pmd);
				DAF.set(pmd);
				modify_eg(Env, Phi2);
				}
			}

		sf_write_double(ouf2, buf2, 2*L);
		sf_write_double(ouf3, buf3, 2*L);		
		sf_write_double(oufx, bufx,   L);

		f2 = 600 - f1 + f2/100;
		f1 = boucle_1.read();
		zz = boucle_2.read();
		pmd.carr_1 = f1;
		pmd.carr_2 = f2;

		if( sometimes(Phi1) )
			{
			recording = !recording;
			DAF.set_C(f1, f2);
			DAF.set_J( zz );
			TAF.set_J( .6*(1-zz) );
			set_period(Env, somecounter);
			if(somecounter & 1)
				mf_triple[i%3] = cf_triple[i%3];
			++somecounter;
			}
		
		if(recording)
			{
			double xi = 2*Phi1.var + .5*(1 + cos(2*M_PI*Phi2.amp));
			boucle_1.overwrite(3*xi + f2/10);
			boucle_2.overwrite( Phi1.zc + Phi2.zc ) ;
			SAF.index(.25 + 9*emm);
			
			if(zz > 0.1)
				nudge_dual(DAF);
			}

		//printf("%3d \t d: %+.2f \t %+.2f, \t t: %+.2f, %+.2f -- f1 %.1f, f2 %.1f\n", 
		//	i, s2.L, s2.R, s3.L, s3.R, f1, f2);
		}
	
	printf("somtimes = %d times \n\n", somecounter);
	sf_close(ouf2);
	sf_close(ouf3);
	sf_close(oufx);
	return 0;
}
