/*
	megaphone
	Copyright Risto Holopainen 2021
	
	Generator for the piece Megaphone


	 This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "phone.h"
#include <cmath>
#include <cstdio>
#include <sndfile.h>

const int sr = 48000;

#define NUMEROSE	// printout of system states over time
#undef NUMEROSE		

class state
{
	public:
		state();
		~state();
		void update();
		void dump(double t);
		void syntbuf(double *buf, int sz, int ch);
		void reset_phase();

	private:
		double h;
		enum {N=15};		// 13 oscillators in original version
		double theta[N];
		double phi[N];
		double delta[N];
		double A[N];
		double psi[N];
		double omega[N];
		double u[N], v[N], a[N];
		double c[N], K, tau;
		double beta[N];
		double Psy;
		int Mu, Mv;
		double g(int);
		double coupling();
		void mu();
		double W;
		int clip;
		double twopion;
		bool dumpheader;
};


state::state()
{
	h = 1.0 / sr;
	W = 2*M_PI/sr;
	twopion = 2*M_PI/N;
	tau = phone::Tau;
	Mu = Mv = 3;
	dumpheader = true;	// print variable names

	for(int i=0; i<N; i++)
		{
		phi[i] = 0.5;
		theta[i] = 0;
		delta[i] = 0;
		A[i] = 1/(i + 1);
		psi[i] = 0;
		a[i] = u[i] = v[i] = c[i] = 1;
		omega[i] = i;
		beta[i] = sqrt(2.0) / (1<<i);
		}
	Psy = 0;
	clip = 0;
}


state::~state()
{
	if(clip)
		printf("%d clipped smp.\n", clip);
}


void state::update()
{	
	K = coupling();
	mu();

	for(int i=0, j; i<N; i++)
		{
		j = (i+3)%N;
		theta[i] += W * (omega[i] + delta[i] - K*sin(theta[i] - phi[j])) ;
		phi[i] 	 += W * (omega[i] - delta[i] - K*sin(phi[i] - theta[j])) ;
		delta[i] += h * (c[i] * fabs(u[i] + v[i]) - delta[i]);

		u[i] = a[i]*sin(theta[i]/Mu);	// output signals
		v[i] = a[i]*sin(phi[i]/Mv);		// output signals

		A[i] += h*tau*(u[i]*u[i] + v[i]*v[i] - .5*A[i] + 0.25*A[j]);
		psi[i] += h * (A[i] - psi[i] * phone::slowness);

		omega[i] = g(i);
		a[i] = cos(psi[i] + i*twopion);
		c[i] = (phone::Three + phone::qwerty*sin(psi[i]));
		Psy += psi[i];
		}
	Psy /= N;
}


double state::g(int j)
{
	double alpha = 1/25.;
	
	for(int i=0, k; i<N; i++)
		{
		k = (i+j)%N;
		if( (A[j] - A[i]) < (A[k] - A[i]) )
			alpha += beta[i];
		}
	if(cos(psi[j]) > 0)
		alpha *= 0.5;

	return alpha * phone::Aleph;
}


double state::coupling()
{
	double k = 0;
	double s = 0;
	
	for(int i=0; i<N; i++)
		s += sin(psi[i]);
	
	k = 1 + s/N;
	return k;
}


void state::mu()
{
	if(sin(Psy) > 1/2.)
		Mu = ( 1 + int(Psy)%5 );

	else if(sin(Psy < -1/2.))
		Mv = ( 2 + int(Psy)%4 );
}


void state::dump(double time)
{
	// printout all
	if(dumpheader)
		{
		printf("t \t ii \t psi \t a \t delta \t omega \t A \n# K: %.2f\n", K);
		dumpheader = false;
		}

	for(int i=0; i<N; i++)
		printf("%.3f \t %2d \t %.4f \t %+.4f \t %.4f \t %7.1f \t %.4f \n",
			time, i, psi[i], a[i], delta[i], omega[i], A[i]);
}


void state::syntbuf(double *buf, int sz, int ch)
{
	// ch : must div by 2
	// output interleaved x,y buf[sz]
	int k = 0;
	double x[ch];
	const double norm = ch * (0.45 / N);
	
	while(k < sz)
		{
		for(int j=0; j<ch; j++)
			x[j] = 0.0;
		this->update();
		for(int i=0; i<N; i += ch/2)
			{
			for(int n=0, j=0; n<ch; n += 2)
				{
				x[n  ] += u[i + j];
				x[n+1] += v[i + j];
				++j;
				}
			}
	
		for(int j=0; j<ch; j++)
			{
			buf[k+j] = norm * x[j];
			if(fabs(buf[k+j]) > 1)
				++clip;
			}
		k += ch;
		}
}


void state::reset_phase()
{
	for(int i=0; i<N; i++)
		{
		theta[i] = 0;
		phi[i] = 0;	
		}
}


int main()
{
	const int ch = 4;
	SNDFILE *ouf;
	SF_INFO head;
	head.samplerate = sr;
	head.channels = ch;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	ouf = sf_open("megaphonE.wav", SFM_WRITE, &head);
	const int L = 1 << 8;
	double buf[L];	
	state s;
	
	for(int i=0; i<64000; i++)	// skipping brief initial transient
		s.update();
	
	s.reset_phase();	
	const int T = phone::duration*sr*ch/L;	// duration 300 sec
	
	for(int i=0; i<T; i++)
		{
		s.syntbuf(buf, L, ch);
		sf_write_double(ouf, buf, L);

		#ifdef NUMEROSE
		if(i%433==1)	// occasional printout
			{
			double tt = (double) i * L / (ch * sr);
			s.dump(tt);
			}
		#endif
		}
	
	sf_close(ouf);
	return 0;
}

