/*
	azcr
	Copyright Risto Holopainen 2021

	adaptive-length zero crossing rate
	finds a fixed number of crossings
	over a variable duration
	
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/


#ifndef AZCR_H
#define AZCR_H


#include <cstdio>
#include <cmath>


class azcr
{
	public: 
		azcr(unsigned int n_crossings=100);
		~azcr();
		double rate(double smp);
		double sd();	// standard-deviation of crossing rate
		
	private:
		unsigned int M;
		double invM;
		unsigned int *delta_t;	// circular buffer of durations between M crossings
		bool cr;				// crossing
		double x1;
		unsigned int i, n;
		unsigned int tspan;
		double eps;
		double zr;				// return value, rate
};


azcr::azcr(unsigned int mm)
{
	M = mm;
	invM = 1.0 / M;
	delta_t = new unsigned int[M];
	x1 = 0;
	i = 0;
	n = 0;
	tspan = 0;		// initiates rate
	eps = 0.001;	// avoid zero division
	
	for(unsigned int i=0; i<M; i++)
		delta_t[i] = 0;
}

azcr::~azcr()
{
	delta_t = 0;
	delete delta_t;
}

double azcr::rate(double x)
{
	cr = (x * x1 < 0);
	x1 = x;
	++n;
	
	if(cr)
		{
		delta_t[i] = n;
		tspan += n;
		n = 0;
		if(++i >= M)
			i = 0;
		tspan -= delta_t[i];
		}

	// avoid zero div. return 0 if too slow
	zr = (tspan > 0) ? (double) (M-1) / tspan : 0;
	return zr;
}


double azcr::sd()
{
	// deviation of delta-t	
	double df = 0.0, sum = 0.0;
	double mean = (double) tspan / M;	// avg period

	for(unsigned int j=0; j<M; j++)
		{
		df = delta_t[j] - mean;
		sum += df*df;
		}
	
	return sqrt(sum/M) / mean;	// st coef of variation ~ [0,1]
}




#endif // def azcr.h
