/*
 *  oscil.h
 *  
 *	Copyright Risto Holopainen 2018
 * 
 *
 * 
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.	
 *
 */

#ifndef OSCIL_H
#define OSCIL_H

#include <cmath>



class oscil
{
	public:
		oscil(int sr, const double *table, int sz, double phase=0.0);	// wavetable
		~oscil();
		double gen(float frq);
		void gen(double *x, int N, float frq);		// to buffer
		double stable(float frq);
	
	private:
		int sr;
		double phase;	// [0,1)
		const double *wavetable;
		int tab_sz;
		double adr;
		int I;		// integer-adr
		double fr;	// fract-adr
		double x;
};



oscil::oscil(int fs, const double *tab, int sz, double ph)
{
// tab[sz] should contain guard point, that is, tab[0] == tab[sz-1]

	sr = fs;
	tab_sz = sz-1;		// one less than tab due to guard point
	wavetable = tab;
	phase = ph;
}




oscil::~oscil()
{
	wavetable = 0;
	delete wavetable;
}



double oscil::gen(float frq)
{
	adr = phase*tab_sz;
	I = (int) adr;
	fr = adr - I;
	x = (1-fr) * wavetable[I] + fr*wavetable[I+1];
	phase += frq/sr;
	
	while(phase >= 1)	// may be slow with insane frequencies!
		phase -= 1;		
	while(phase < 0)
		phase += 1;

	return x;
}


void oscil::gen(double *s, int N, float frq)
{
	double omega = frq/sr;
	for(int j=0; j<N; j++)
		{
		adr = phase*tab_sz;
		I = (int) adr;
		fr = adr - I;
		x = fr*wavetable[I] + (1-fr)*wavetable[I+1];	
		s[j] = x;

		phase += omega;	
		if(phase >= 1)		// be careful, avoid |frq| > sr ! 
			phase -= 1;		
		else if(phase < 0)
			phase += 1;
		}
}


double oscil::stable(float frq)
{
	adr = phase*tab_sz;
	I = (int) adr;
	fr = adr - I;
	x = (1-fr) * wavetable[I] + fr*wavetable[I+1];
	phase += frq/sr;
	
	if(phase >= 1.0 || phase < 0.0)
		phase -= floor(phase);		

	return x;	
}

#endif
