/*
 *  timedomain.h
 *  a few simple time domain feature extractors. 
 *
 *  Copyright Risto Holopainen on 2011 and 2021
 
 
	This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */


#include "delay_t.h"
#include <cmath>
#include <vector>

using std::vector;


/*
	zcr
	extrema
	halfwave
	RMS: 1-pole
	+ boxcar?
	tv: total variation

	---TODO---
	*centroid
	*a1a3
	*fluxus --- filterbank flux, const.Q !! or not.
	(CQT --> frq-dependent delay times in flux.)
	
	---
	avg, sd, of above measures
	+ entropy of countables (zcr, extrema)
*/



class zcr
{
	public:
		zcr(int L=800);	// period i smp att integrera över
		~zcr() { delete D;}
		float operator()(double);
		float rate(const double *smp, int L);
		float robust(double);	// tolerate low-amp noise
		void set_eps(double E = 1E-3) {eps = E;} // default for resetting
		
	private:
		int L;
		delay<double> *D;
		double xd, x1, xe;
		int count, flag;
		float avg;
		double eps;
};

zcr::zcr(int p)
{
	// count zero crossings per L smp

	L = p;
	D = new delay<double>(L, 0.0);	
	avg = 1.0/L;
	count = 0;
	x1 = xe = 0;
	eps = 1E-3;	// for robust detection
}




float zcr::operator()(double x)
{
	xd = D->z(x);	// pure delay.
	if( (x > 0 && x1 < 0) || (x < 0 && x1 > 0) )
		++count;

	// simplify !!
	if( (xd > 0 && xe < 0) || (xd < 0 && xe > 0) )
		--count;

	x1 = x;
	xe = xd;

	return count * avg;	
}

float zcr::robust(double x)
{
	// detect zc iff distance from 0 > eps

	flag = 0;

	if( (x > eps && x1 < -eps)
		|| (x < -eps && x1 > eps) )
		{
		++count;
		flag = 1;
		}

	x1 = x;
	if(D->z(flag))
		--count;

	return count * avg;	
}



float zcr::rate(const double *smp, int L)
{
	count = 0;
	for(int i=0; i<L-1; i++)
		if(smp[i+1] > 0 && smp[i] < 0)
			++count;
	return count*2.0/L;
}


class tv
{
// total variation over last sz samples.
	public:
		tv(int sz);
		~tv() { DL=0; delete DL; }
		double local(double x);
		
	private:
		delay<double> *DL;
		int N;
		double v, x1, s, norm;
};

tv::tv(int sz)
{
	N = sz;
	DL = new delay<double>(N, 0.0);
	v = x1 = s = 0;
	norm = 1.0 / N;
}

double tv::local(double x)
{
	v = fabs(x-x1);
	s += v;
	s -= DL->z(v);	
	x1 = x;
	return norm*s;
}



class extrema
{
// relative frequency of maxima per window in [0,1]
// do not use member functions in parallel !
	public:
		extrema(unsigned int sz, int ord=2);
		~extrema();
		float max(double x);			// or peak-count
		float max(double x, int *p);	// p[ord]: inter-peak delta-t
		//float min(double x) == max(-x); !!
		double robust(double x);		// : peak must be noticeably greater
		void set_epsilon(double ee) { epsilon = (ee >= 0.0) ? ee : 0; }
		
	private:
		unsigned int L;
		unsigned int MC;
		delay<double> *DL;
		double u[2], flag;
		double norm;
		unsigned long t;
		int *P;			// table of delta-t's of peaks
		int N, dt, k;	// N: order
		double epsilon;	// tolerance in robust.
};



extrema::extrema(unsigned int sz, int ord)
{
	L = sz;
	MC = 0;
	DL = new delay<double>(L, 0.0);
	u[0] = u[1] = 0;
	norm = 2.0 / (L+1);	// range: [0,1]
	t = 0;
	N = ord;
	P = new int[N];
	dt = 0;
	k = 0;
	epsilon = 0.001;
}


extrema::~extrema()
{
	DL = 0;
	P = 0;
	delete DL;
	delete P;
}


float extrema::max(double x)
{
	// count maxima
	flag = 0;
	if(u[1] < u[0] && u[0] > x)
		{
		++MC;
		flag = 1;
		}

	if(DL->z(flag))
		--MC;

	u[1] = u[0];
	u[0] = x;
	
	return (float) norm*MC;
}


float extrema::max(double x, int *pp)
{
// return 2x #maxima; send other stats to p

	flag = 0;
	if(u[1] <= u[0] && u[0] > x)	// halvplatå.
		{
		++MC;
		flag = 1;
		P[k] = t-dt;
		dt = t;
		for(int j=0; j<N; j++)
			pp[(k+j)%N] = P[j];
		// copy elementwise P --> pp
		++k %= N;
		}

	if(DL->z(flag))
		--MC;

	u[1] = u[0];
	u[0] = x;
	++t;
	
	return (float) norm*MC;
}



double extrema::robust(double x)
{
	// count maxima if more than epsilon above neighbours
	flag = 0;
	if(u[1] + epsilon < u[0] 
		&& u[0] > x + epsilon)
		{
		++MC;
		flag = 1;
		}

	if(DL->z(flag))
		--MC;

	u[1] = u[0];
	u[0] = x;
	return (double) norm*MC;
}



class halfwave
{
// half-wave-rectifier : sample-wise update
// OK.
	public:
		halfwave(unsigned int sz);
		~halfwave();
		double amp(double x);
		
	private:
		unsigned int L;
		double A, h;
		double norm;
		delay<double> *DL;
};


halfwave::halfwave(unsigned int sz)
{
	L = sz;
	DL = new delay<double>(L, 0.0);
	A = 0;
	norm = 1.0 / L;	
}

halfwave::~halfwave()
{
	DL = 0;
	delete DL;
}



double halfwave::amp(double x)
{
	h = (x > 0) ? norm*x : 0;
	A += h - DL->z(h);
	return A;
}



class fullwave
{
	public:
		fullwave(int sz);
		~fullwave() { DL=0; delete DL; }
		double amp(double x);
		
	private:
		delay<double> *DL;
		int N;
		double A, u, norm;
};


fullwave::fullwave(int sz)
{
	N = sz;
	DL = new delay<double>(N, 0.0);
	A = 0;
	norm = 1.0 / N;
}



double fullwave::amp(double x)
{
	u = norm*fabs(x);
	A += u - DL->z(u);
	return A;
}



class rms
{
// using 1-pole leaky integrator

	public:
		rms(float time_const = 1000);	// #smp decay time
		double env(double x);
		void set_b(float tc) {b = exp(-1/tc);}

		
	private:
		double b, eps;
		double y, y1;
};


rms::rms(float tc)
{
// suppose b = exp(-w_c), w_c = 1/tc

	b = exp(-1/tc);
	y = 0;
	eps = 1E-25;
}


double rms::env(double x)
{
	y = (1-b)*x*x + b*y;
	return sqrt(y + eps);
}








